;;; -*- lisp -*-
;;; -- components.lisp --

(in-package :cl-webobjects)

(defgeneric cform-rewinding-p (obj page-obj)
  (:documentation "Internal method to determine, during the rewinding phase, if the COMP has been fired for calling its action.
- OBJ the wocomponent instance
- PAGE-OBJ the wocomponent owner page"))
 
;--------------------------------------------------------------------------------

(defcomponent cform () ()
	      (:documentation "This component render as a FORM tag class, but it is aware of
the request cycle and is able to fire an action on rewind"))

(defmethod cform-rewinding-p ((obj cform) (pobj page))
  (string= (htcomponent-client-id obj)
	   (page-req-parameter pobj *rewind-parameter*)))

(defmethod wocomponent-parameters ((o cform))
  (list :id :required :action nil))

(defmethod wocomponent-template((o cform))
  (let ((client-id (htcomponent-client-id o)))
    (when (null client-id)
      (setf client-id ""))
    (form> :static-id client-id
	   :name client-id
	   (wocomponent-informal-parameters o)
	   (input> :name *rewind-parameter*
		   :type "hidden" 		 
		   :value client-id)
	   (htcomponent-body o))))

(defmethod wocomponent-before-rewind ((obj cform) (pobj page))
  (setf (page-current-form pobj) obj))

(defmethod wocomponent-after-rewind ((obj cform) (pobj page))
  (let ((action (wocomponent-parameter-value obj :action)))
    (unless (or (null action) (null (cform-rewinding-p obj pobj))) 
      (funcall (fdefinition action) pobj))
    (setf (page-current-form pobj) nil)))

;--------------------------------------------------------------------------------

(defcomponent action-link (cform) ()
	      (:documentation "This component behaves like a CFORM, firing it's associated action once clicked.
It renders as a normal link."))

(defmethod wocomponent-reserved-parameters ((o action-link))
  '(:href))

(defmethod wocomponent-template((o action-link))
  (let ((client-id (htcomponent-client-id o)))
    (when (null client-id)
      (setf client-id ""))
    (a> :static-id client-id
	:href (format nil "?~a=~a" *rewind-parameter* client-id)
	(wocomponent-informal-parameters o)
	(htcomponent-body o))))

;---------------------------------------------------------------------------------------

(defcomponent cinput ()
    ((result-as-list :initarg :result-as-list
		     :accessor cinput-result-as-list))
    (:default-initargs :result-as-list nil)
    (:documentation "Request cycle aware component the renders as an INPUT tag class"))

(defmethod wocomponent-parameters ((o cinput))
  (list :id :required :reader nil :writer nil :visit-object nil :accessor nil :type :required))

(defmethod wocomponent-reserved-parameters ((o cinput))
  '(:value :name))

(defmethod wocomponent-template ((obj cinput))
  (let ((client-id (htcomponent-client-id obj))
	(type (wocomponent-parameter-value obj :type))
	(visit-object (wocomponent-parameter-value obj :visit-object))
	(accessor (wocomponent-parameter-value obj :accessor))
	(reader (wocomponent-parameter-value obj :reader))
	(value ""))
    (when (null visit-object)
      (setf visit-object (htcomponent-page obj)))
    (if (and (null reader) accessor)		  
	(setf value (funcall (fdefinition accessor) visit-object))
	(setf value (funcall (fdefinition reader) visit-object)))
    (input> :static-id client-id
	    :type type
	    :name client-id
	    :value value
	    (wocomponent-informal-parameters obj))))

(defmethod wocomponent-after-rewind ((obj cinput) (pobj page))
  (let ((visit-object (wocomponent-parameter-value obj :visit-object))
	(accessor (wocomponent-parameter-value obj :accessor))
	(writer (wocomponent-parameter-value obj :writer))	  
	(new-value (page-req-parameter pobj 
				       (htcomponent-client-id obj)
				       (cinput-result-as-list obj))))
    (unless (null new-value)
      (when (null visit-object)
	(setf visit-object (htcomponent-page obj)))
      (if (and (null writer) accessor)
	  (funcall (fdefinition `(setf ,accessor)) new-value visit-object)
	  (funcall (fdefinition writer) new-value visit-object)))))

;---------------------------------------------------------------------------------------
(defcomponent csubmit () ()
	      (:documentation "This component render as an INPUT tag class ot type submit, but 
can override the default CFORM action, using its own associated action"))

(defmethod wocomponent-parameters ((o csubmit))
  (list :id :required :value :required :action nil))

(defmethod wocomponent-reserved-parameters ((o csubmit))
  '(:type :name))

(defmethod wocomponent-template ((obj csubmit))
  (let ((client-id (htcomponent-client-id obj))
	(value (wocomponent-parameter-value obj :value)))
    (input> :static-id client-id
	    :type "submit"
	    :name client-id
	    :value value
	    (wocomponent-informal-parameters obj))))

(defmethod wocomponent-after-rewind ((obj csubmit) (pobj page))
  (let ((action (wocomponent-parameter-value obj :action))
	(current-form (page-current-form pobj))
	(submitted-p (page-req-parameter pobj (htcomponent-client-id obj))))
    (unless (or (null current-form) (null submitted-p) (null action))
      (setf (getf (wocomponent-parameters current-form) :action) action))))
;-----------------------------------------------------------------------------
(defcomponent submit-link (csubmit) ()
	      (:documentation "This component renders as a normal link, but behaves like a CSUBMIT,
so it can be used instead of CSUBMIT when needed"))

(defmethod wocomponent-reserved-parameters ((o submit-link))
  '(:href))

(defmethod wocomponent-template ((obj submit-link))
  (let* ((id (htcomponent-client-id obj))
	 (submit-id (generate-id id)))
    (list 
     (input> :static-id submit-id
	     :style "display:none;"
	     :type "submit"
	     :name id
	     :value "-")
     (a> :static-id id
	:href (format nil "javascript:document.getElementById('~a').click();" submit-id)
	(wocomponent-informal-parameters obj)
	(htcomponent-body obj)))))

;--------------------------------------------------------------------------

(defcomponent cselect (cinput) ()
	      (:default-initargs :result-as-list t)
	      (:documentation "This component renders as a normal SELECT tag class, 
but it is request cycle aware."))

(defmethod wocomponent-parameters :around ((obj cselect))
  (declare (ignore obj))
  (let ((params (call-next-method)))
    (remf params :reader)
    (remf params :type)
    params))

(defmethod wocomponent-reserved-parameters ((obj cselect))
  (declare (ignore obj))
  '(:type :name))

(defmethod wocomponent-template ((obj cselect))
  (let ((client-id (htcomponent-client-id obj)))
    (select> :static-id client-id
	     :name client-id
	     (wocomponent-informal-parameters obj)
	     (htcomponent-body obj))))

