;; -*- lisp -*-

(in-package :cl-webobjects)

(defgeneric woserver-register-lisplet (obj lisplet-obj)
  (:documentation "This method registers a lisplet for request dispatching
- OBJ the WOSERVER instance
- LISPLET-OBJ the LISPLET instance"))

(defgeneric woserver-unregister-lisplet (obj lisplet-obj)
  (:documentation "This method unregisters a lisplet from request dispatching
- OBJ the WOSERVER instance
- LISPLET-OBJ the LISPLET instance"))

(defgeneric woserver-dispatch-request (obj)) ;internal
(defgeneric woserver-dispatch-method (obj)) ;internal

(defgeneric woserver-start (obj)
  (:documentation "Starts the server"))
(defgeneric woserver-stop (obj)
  (:documentation "Stops the server"))

(defgeneric (setf woserver-port) (val obj))
(defgeneric (setf woserver-sslport) (val obj))
(defgeneric (setf woserver-address) (val obj))
(defgeneric (setf woserver-name) (val obj))
(defgeneric (setf woserver-sslname) (val obj))
(defgeneric (setf woserver-mod-lisp-p) (val obj))
(defgeneric (setf woserver-use-apache-log-p) (val obj))
(defgeneric (setf woserver-input-chunking-p) (val obj))
(defgeneric (setf woserver-read-timeout) (val obj))
(defgeneric (setf woserver-write-timeout) (val obj))
#+(and :unix (not :win32)) (defgeneric (setf woserver-setuid) (val obj))
#+(and :unix (not :win32)) (defgeneric (setf woserver-setgid) (val obj))
#-:hunchentoot-no-ssl (defgeneric (setf woserver-ssl-certificate-file) (val obj))
#-:hunchentoot-no-ssl (defgeneric (setf woserver-ssl-privatekey-file) (val obj))
#-:hunchentoot-no-ssl (defgeneric (setf woserver-ssl-privatekey-password) (val obj))


(defclass page404 (page) 
  ((style :initform 
	  "
body {
  font-family: arial, elvetica;
  font-size: 7pt;
}
span.blue {
  background-color: #525D76;
  color: white;
  font-weight: bolder;  
  margin-right: .25em;
}
p.h1, p.h2 {
  background-color: #525D76;
  color: white;
  font-weight: bolder;
  font-size: 2em;
  margin: 0;
  margin-bottom: .5em;
}
p.h2 {font-size: 1.5em;}"
	  :reader page404-style))
  (:documentation "This page class is used to render 
the 404 (page not found) messages."))

(defmethod page-content ((obj page404))
  (html>
   (head>
    (title>
     "404 Page not found")
    (style> 
     (page404-style obj)))
   (body>
    (p>
     (p> :class "h1"
	 (format nil "HTTP Status 404 - ~a" (request-uri *request*)))
     (hr> :noshade "noshade")
     (p>
      (span> :class "blue"
	     ($> "type"))
      "Status report")
     (p>
      (span> :class "blue"
	     "message")
      (request-uri *request*))
     (p>
      (span> :class "blue"
	     "description")
      (format nil "The requested resource (~a) is not available." (request-uri *request*)))
     (hr> :noshade "noshade"))
    (p> :class "h2"
	"cl-webobject server"))))

(defclass woserver ()
  ((port :initarg :port
	 :reader woserver-port)
   (sslport :initarg :sslport
	 :reader woserver-sslport)
   (address :initarg :address
	    :reader woserver-address)
   (name :initarg :name
	 :reader woserver-name)
   (sslname :initarg :sslname
	 :reader woserver-sslname)
   (mod-lisp-p :initarg :mod-lisp-p
	       :reader woserver-mod-lisp-p)
   (use-apache-log-p :initarg :use-apache-log-p
		     :reader woserver-use-apache-log-p)
   (input-chunking-p :initarg :input-chunking-p
		     :reader woserver-input-chunking-p)
   (read-timeout :initarg :read-timeout
		 :reader woserver-read-timeout)
   (write-timeout :initarg :write-timeout
		 :reader woserver-write-timeout)
   #+(and :unix (not :win32)) (setuid :initarg :setuid
				      :reader woserver-setuid)
   #+(and :unix (not :win32)) (setgid :initarg :setgid
				      :reader woserver-setgid)
   #-:hunchentoot-no-ssl (ssl-certificate-file :initarg :ssl-certificate-file
					       :reader woserver-ssl-certificate-file)
   #-:hunchentoot-no-ssl (ssl-privatekey-file :initarg :ssl-privatekey-file
					       :reader woserver-ssl-privatekey-file)
   #-:hunchentoot-no-ssl (ssl-privatekey-password :initarg :ssl-privatekey-password
					       :reader woserver-ssl-privatekey-password)
   (server :initform nil
	   :accessor woserver-server)
   (sslserver :initform nil
	   :accessor woserver-sslserver)
   (lisplets :initform (make-hash-table :test 'equal)
	     :reader woserver-lisplets)
   (page404 :initarg :page404
	    :accessor woserver-page404))
  (:default-initargs :address nil 
    :name (gensym)
    :sslname (gensym)
    :port 80 
    :sslport 443
    :mod-lisp-p nil
    :input-chunking-p t
    :read-timeout *default-read-timeout* 
    :write-timeout *default-write-timeout*
    #+(and :unix (not :win32)) :setuid nil
    #+(and :unix (not :win32)) :setgid nil
    #-:hunchentoot-no-ssl :ssl-certificate-file nil
    #-:hunchentoot-no-ssl :ssl-privatekey-password nil
    :page404 (make-instance 'page404))
  (:documentation "WOSERVER is built around huncentoot and has the 
instructions for lisplet dispatching, so use this class to start and stop 
hunchentoot server."))

(defmethod initialize-instance :after ((obj woserver) &rest keys)
  (let ((use-apache-log-p (getf keys :use-apache-log-p :undefined))
	#-:hunchentoot-no-ssl (ssl-privatekey-file (getf keys :ssl-privatekey-file :undefined)))
    (when (eq use-apache-log-p :undefined)
      (setf (woserver-use-apache-log-p obj) (getf keys :mod-lisp-p)))
    #-:hunchentoot-no-ssl (when (eq ssl-privatekey-file :undefined)
			    (setf (woserver-ssl-privatekey-file obj) (getf keys :ssl-certificate-file)))))
      
(defmethod woserver-register-lisplet ((obj woserver) (lisplet-obj lisplet))
  (let ((lisplets (woserver-lisplets obj))
	(server-base-path *woserver-base-path*)
	(location (lisplet-base-path lisplet-obj)))
    (unless (null server-base-path)
      (setf location (format nil "~@[~a~]~a/" server-base-path location)))
    (setf (gethash location lisplets) (create-prefix-dispatcher 
				    location
				    #'(lambda ()										
					(lisplet-dispatch-method lisplet-obj)))))) 

(defmethod woserver-unregister-lisplet ((obj woserver) (lisplet-obj lisplet))
  (let ((lisplets (woserver-lisplets obj))
	(location (lisplet-base-path lisplet-obj)))
    (remhash location lisplets))) 


;;;-------------------------- WRITERS ----------------------------------------

(defmethod (setf woserver-port) (val (obj woserver))
  (unless (null (woserver-server obj))
    (error "Cannot change port when server is started"))
  (setf (slot-value obj 'port) val))

(defmethod (setf woserver-sslport) (val (obj woserver))
  (unless (null (woserver-server obj))
    (error "Cannot change SSL port when server is started"))
  (setf (slot-value obj 'sslport) val))

(defmethod (setf woserver-address) (val (obj woserver))
  (unless (null (woserver-server obj))
    (error "Cannot change binding address when server is started"))
  (setf (slot-value obj 'address) val))

(defmethod (setf woserver-name) (val (obj woserver))
  (unless (null (woserver-server obj))
    (setf (server-name (woserver-server obj)) val))
  (setf (slot-value obj 'name) val))

(defmethod (setf woserver-sslname) (val (obj woserver))
  (unless (null (woserver-sslserver obj))
    (setf (server-name (woserver-sslserver obj)) val))
  (setf (slot-value obj 'sslname) val))

(defmethod (setf woserver-mod-lisp-p) (val (obj woserver))
  (unless (null (woserver-server obj))
    (error "Cannot change mod-lisp property when server is started"))
  (setf (slot-value obj 'mod-lisp-p) val))

(defmethod (setf woserver-use-apache-log-p) (val (obj woserver))
  (unless (null (woserver-server obj))
    (error "Cannot change logging property when server is started"))
  (setf (slot-value obj 'use-apache-log-p) val))

(defmethod (setf woserver-input-chunking-p) (val (obj woserver))
  (unless (null (woserver-server obj))
    (error "Cannot change chunking property when server is started"))
  (setf (slot-value obj 'input-chunking-p) val))

(defmethod (setf woserver-read-timeout) (val (obj woserver))
  (unless (null (woserver-server obj))
    (error "Cannot change read timeout property when server is started"))
  (setf (slot-value obj 'read-timeout) val))

(defmethod (setf woserver-write-timeout) (val (obj woserver))
  (unless (null (woserver-server obj))
    (error "Cannot change write timeout property when server is started"))
  (setf (slot-value obj 'write-timeout) val))

#+(and :unix (not :win32)) (defmethod (setf woserver-setuid) (val (obj woserver))
			     (unless (null (woserver-server obj))
			       (error "Cannot change uid property when server is started"))
			     (setf (slot-value obj 'setuid) val))

#+(and :unix (not :win32)) (defmethod (setf woserver-setgid) (val (obj woserver))
			     (unless (null (woserver-server obj))
			       (error "Cannot change gid property when server is started"))
			     (setf (slot-value obj 'setgid) val))

#-:hunchentoot-no-ssl (defmethod (setf woserver-ssl-certificate-file) (val (obj woserver))
			(unless (null (woserver-server obj))
			  (error "Cannot change ssl certificate file property when server is started"))
			(setf (slot-value obj 'ssl-certificate-file) val))

#-:hunchentoot-no-ssl (defmethod (setf woserver-ssl-privatekey-file) (val (obj woserver))
			(unless (null (woserver-server obj))
			  (error "Cannot change ssl privatekey file property when server is started"))
			(setf (slot-value obj 'ssl-privatekey-file) val))

#-:hunchentoot-no-ssl (defmethod (setf woserver-ssl-privatekey-password) (val (obj woserver))
			(unless (null (woserver-server obj))
			  (error "Cannot change ssl privatekey password property when server is started"))
			(setf (slot-value obj 'ssl-privatekey-password) val))

;;;-------------------------- METHODS ----------------------------------------
(defmethod woserver-dispatch-request ((obj woserver))
  (let ((lisplets (woserver-lisplets obj)))
    (loop for dispatcher being the hash-value of lisplets
       for action = (funcall dispatcher *request*)
       when action return (funcall action))))


(defmethod woserver-dispatch-method ((obj woserver))
  (let ((page404 (woserver-page404 obj))
	(result nil))
    (progn 
      (setf result (woserver-dispatch-request obj)) 
      (if (null result)
	#'(lambda () (with-output-to-string (*standard-output*) (page-render page404)))
	#'(lambda () result)))))

(defmethod woserver-start ((obj woserver))
  (let ((port (woserver-port obj))
	(sslport (woserver-sslport obj))
	(address (woserver-address obj))
	(dispatch-table (list #'(lambda (request) 
						  (declare (ignorable request))
						  (woserver-dispatch-method obj))))
	(name (woserver-name obj))
	(sslname (woserver-sslname obj))
	(mod-lisp-p (woserver-mod-lisp-p obj))
	(use-apache-log-p (woserver-use-apache-log-p obj))
	(input-chunking-p (woserver-input-chunking-p obj))
	(read-timeout (woserver-read-timeout obj))
	(write-timeout (woserver-write-timeout obj))
	(uid (woserver-setuid obj))
	(gid (woserver-setgid obj))
	(ssl-certificate-file (woserver-ssl-certificate-file obj))
	(ssl-privatekey-file (woserver-ssl-privatekey-file obj))
	(ssl-privatekey-password (woserver-ssl-privatekey-password obj)))
    (progn
      (setf (woserver-server obj)
	    (start-server :port port
			  :address address
			  :dispatch-table dispatch-table
			  :name name
			  :mod-lisp-p mod-lisp-p
			  :use-apache-log-p use-apache-log-p
			  :input-chunking-p input-chunking-p
			  :read-timeout read-timeout
			  :write-timeout write-timeout
			  #+(and :unix (not :win32)) :setuid uid
			  #+(and :unix (not :win32)) :setgid gid))
      #-:hunchentoot-no-ssl (when ssl-certificate-file
			      (setf (woserver-sslserver obj)
				    (start-server :port sslport
						  :address address
						  :dispatch-table dispatch-table
						  :name sslname
						  :mod-lisp-p mod-lisp-p
						  :use-apache-log-p use-apache-log-p
						  :input-chunking-p input-chunking-p
						  :read-timeout read-timeout
						:write-timeout write-timeout
						#+(and :unix (not :win32)) :setuid uid
						#+(and :unix (not :win32)) :setgid gid
						:ssl-certificate-file ssl-certificate-file
						:ssl-privatekey-file ssl-privatekey-file
						:ssl-privatekey-password ssl-privatekey-password))))))
  
(defmethod woserver-stop ((obj woserver))
  (progn 
    (setf (woserver-server obj) (stop-server (woserver-server obj)))
    (when (woserver-sslserver obj)
      (setf (woserver-sslserver obj) (stop-server (woserver-sslserver obj))))))
;;;----------------------------------------------------------------------------
(defun start-woserver (woserver-obj 
			&key (port 80)		       
			address
			(name (gensym))
			(mod-lisp-p nil)
			(use-apache-log-p mod-lisp-p)
			(input-chunking-p t)
			(read-timeout *default-read-timeout*)
			(write-timeout *default-write-timeout*)
			#+(and :unix (not :win32)) setuid
			#+(and :unix (not :win32)) setgid
			#-:hunchentoot-no-ssl ssl-certificate-file
			#-:hunchentoot-no-ssl (ssl-privatekey-file ssl-certificate-file)
			#-:hunchentoot-no-ssl ssl-privatekey-password)
	(start-server :port port
		      :address address
		      :dispatch-table (list #'(lambda (request) 
						(declare (ignorable request))
						(woserver-dispatch-method woserver-obj)))
		      :name name
		      :mod-lisp-p mod-lisp-p
		      :use-apache-log-p use-apache-log-p
		      :input-chunking-p input-chunking-p
		      :read-timeout read-timeout
		      :write-timeout write-timeout
		      #+(and :unix (not :win32)) :setuid setuid
		      #+(and :unix (not :win32)) :setgid setgid
		      #-:hunchentoot-no-ssl :ssl-certificate-file ssl-certificate-file
		      #-:hunchentoot-no-ssl :ssl-privatekey-file ssl-privatekey-file
		      #-:hunchentoot-no-ssl :ssl-privatekey-password ssl-privatekey-password))
			      
  